import xesam

import xml.sax
from xml.sax.saxutils import escape
from cStringIO import StringIO

ITEM_TEMPLATE = \
"""[%s]
Parents=%s
Children=%s
%s
"""

XESAM_URL_SOURCE = "xesam:Source"
XESAM_DATA_OBJECT = "xesam:DataObject"

class OntologyError (Exception):
	def __init__ (self, msg=None, cause=None):
		if msg:
			Exception.__init__ (self, msg)
		else:
			Exception.__init__ (self)

class Ontology :
	"""
	An abstract in-memory representation of an ontology.
	@param roots: the root items of the ontology
	"""
	def __init__ (self, roots):
		self._roots = roots
		self._contents = {}
		self._fields = {}
		self._sources = {}
		
		# Special case DataObject since it can contain both Source and Content
		# it is just an abstract base class
		extra_roots = []
		for root in roots:
			if root.get_id() == XESAM_DATA_OBJECT:
				extra_roots.extend(root.get_all_children())
		roots.extend(extra_roots)
		
		for root in roots:
			if root.get_id() == XESAM_DATA_OBJECT:
				continue
			if root.get_type() == Item.TYPE_CONTENT:
				self._contents.update([ (item.get_id(), item) for item in root.get_all_children()])
				self._contents[root.get_id()] = root
			elif root.get_type() == Item.TYPE_SOURCE:
				self._sources.update ([ (item.get_id(), item) for item in root.get_all_children()])
				self._sources[root.get_id()] = root
			elif root.get_type() == Item.TYPE_FIELD:
				self._fields.update ([ (item.get_id(), item) for item in root.get_all_children()])
				self._fields[root.get_id()] = root
			else:
				xesam.error ("Unknown item type for %s" % root)
		
		# Add all fields defined by content types to the field index
		for cont in self._contents.itervalues():
			self._fields.update([(field.get_id(), field) for field in cont.get_all_fields()])
		
		# Add all fields defined by the source type to the field index
		for src in self._sources.itervalues():
			self._fields.update([(field.get_id(), field) for field in src.get_all_fields()])
	
	def has_field (self, field_name):
		return self._fields.has_key(field_name)
		
	def has_content (self, content_name):
		return self._contents.has_key(content_name)
	
	def has_source (self, source_name):
		return self._sources.has_key(source_name)
	
	def get_item (self, item_id):
		
		if self.has_field(item_id) : return self._fields[item_id]
		elif self.has_content(item_id) : return self._contents[item_id]
		elif self.has_source(item_id) : return self._sources[item_id]
		else:
			raise KeyError (item_id)
		
	def get_roots (self):
		return self._roots
	
	def get_categories (self):
		"""
		Return all sources and content types defined in this ontology
		"""
		return list(self._sources.itervalues()) + list(self._contents.itervalues())
	
	def get_fields (self):
		return self._fields.itervalues()
	
	def get_contents (self):
		return self._contents.itervalues()
	
	def get_sources (self):
		return self._sources.itervalues()
	
	def get_fields (self):
		"""
		Return all fields defined by this ontology
		"""
		return self._fields.itervalues()
	
class Item :	
	"""
	An abstract item described by an ontology. You can access
	the item properties as if the object where a dict. Iterating
	over an Item object will iterate over all property names of
	the Item.
	
	The actual item implementation goes together with the ontology
	implementation.
	"""

	[
		TYPE_FIELD,
		TYPE_CONTENT,
		TYPE_SOURCE
	] = range (3)
	

	def __init__ (self, item_id=None, item_type=TYPE_FIELD, external=False):
		self._props = {}
		self._parents = {}
		self._children = {}
		self._id = item_id
		self._type = item_type
		self._external = external # Whether or not the Item comes from an external onto
	
	def __getitem__ (self, name):
		return self._props[name]

	def __setitem__ (self, name, value):
		self._props[name] = value
	
	def __iter__ (self):
		return self._props.iterkeys()
	
	def __str__ (self):
		"""
		Returns as a .ini-like represenatation og the item
		"""
		return ITEM_TEMPLATE % \
		(self._id,
		", ".join([parent.get_id() for parent in self.get_parents()]),
		", ".join (self.get_child_names()),
		"\n".join (["%s=%s" % (prop, self._props[prop]) for prop in self._props.iterkeys()]))
	
	def get_properties (self):
		return self._props
		
	def get_id (self):
		"""
		Return unique name for this items
		"""
		return self._id
	
	def set_id (self, item_id):
		self._id = item_id
		
	def get_type (self):
		"""
		Return the type of this Item. It is one of the
		Item.TYPE_* enumeration.
		"""
		return self._type
	
	def set_type (self, item_type):
		self._type = int(item_type)
	
	
	def get_parents (self):
		if len(self._parents) == 0:
			root = RootItem()
			return [root]
		return self._parents.itervalues()
	
	def get_all_parents (self):
		"""
		Recursively look up all parent items.
		"""
		parents = []
		for parent in self.get_parents():
			# Skip external children
			if parent.is_external() : continue 
			
			parents.append(parent)
			parents.extend(parent.get_all_parents())
		return parents
	
	def add_parent (self, parent, notify_parent=True):
		if parent is None:
			xesam.debug("Warning: Tried to add None parent to " + self.get_id())
			return
		if not isinstance (parent, Item):
			xesam.error ("Adding non-Item type '%s' as parent" % type(parent))
			return
		self._parents[parent.get_id()] = parent
		if notify_parent:
			parent.add_child (self, notify_child=False)
	
	def add_child (self, child, notify_child=True):
		"""
		This method should only be called from inside Item.set_parent()
		"""
		self._children[child.get_id()] = child
		if notify_child:
			child.add_parent(self, notify_parent=False)
	
	def remove_child (self, child):
		"""
		This method should only be called from inside Item.set_parent()
		"""
		self._children.remove(child)
	
	def get_children (self):
		return self._children.itervalues()
	
	def num_children (self):
		return len(self._children)
	
	def get_child_names (self):
		return self._children.iterkeys()
	
	def get_all_children (self):
		"""
		Recursively look up all child items. This will not traverse into branches
		with external items.
		"""
		children = []
		for child in self.get_children():
			# Skip external children
			if child.is_external() : continue 
			
			children.append(child)
			children.extend(child.get_all_children())
		return children
	
	def get_all_child_names (self):
		"""
		Recursively fetch names of all child items. This will not traverse into branches
		with external children
		"""
		children = []
		for child in self.get_children():
			# Skip external children
			if child.is_external() : continue
			
			children.append(child.get_id())
			children.extend(child.get_all_child_names())
		return children
	
	def is_child_of (self, item):
		"""
		@param item: a xesam.ontology.Item or and Item id as return by Item.get_id()
		@return : true if this item is a child of the item argument
		"""
		item_id = item
		if isinstance (item, Item):
			item_id = item.get_id()
		
		if isinstance(self, RootItem) and item != self.get_id():
			return False
		
		if self._id == item_id :
			return True
				
		for parent in self.get_parents():
			if parent.is_child_of (item):
				return True
		return False
	
	def add_properties (self, props):
		"""
		@param props: a dict of prop->value pairs
		"""
		self._props.update(props)
		
	def is_external (self):
		"""
		Return whether this item is resolved against an external ontology - 
		ie an ontology that should only be cross reffed for consistensy checking
		"""
		return self._external
		
	def set_external (self, external):
		"""
		Set boolean value whether or not this item should be regarded as external
		"""
		self._external = external
		
class RootItem (Item):
	def __init__ (self, item_type=Item.TYPE_FIELD):
		Item.__init__ (self, "Root Item", item_type)

class Field (Item):

	DATA_TYPE_STRING = "string"
	DATA_TYPE_BOOLEAN = "boolean"
	DATA_TYPE_FLOAT = "float"
	DATA_TYPE_INT = "int"
	DATA_TYPE_DATE = "data"
	
	DATA_TYPES =\
	[
		DATA_TYPE_STRING,
		DATA_TYPE_BOOLEAN,
		DATA_TYPE_FLOAT,
		DATA_TYPE_INT,
		DATA_TYPE_DATE
	]
	"""A list of valid data types"""

	def __init__ (self, item_id, external=False, data_type=DATA_TYPE_STRING):
		Item.__init__ (self, item_id, item_type=Item.TYPE_FIELD, external=external)
		self._categories = []
		self._data_type = data_type
		
	def add_category (self, cat, notify_parent=True):
		if not isinstance (cat, FieldMapper):
			xesam.error ("Adding illegal category of type '%s' to field %s" % (type(cat), self.get_id()))
		self._categories.append(cat)
		if notify_parent:
			cat.add_field (self, notify_child=False)
	
	def get_categories (self):
		return self._categories
	
	def is_abstract (self):
		"""
		Returns true iff this field has children
		"""
		return len(self._children) != 0 and self._data_type is None
	
	def is_in_map (self, cat):
		"""
		@param cat: a xesam.ontology.FieldMapper or an id obtained via get_id() on a FieldMapper
		@return : True if the category implies this field
		"""
		raise SystemExit ("FU")
		if isinstance (cat, Content):
			cat = cat.get_id()
		
		parent_cat = get_category()
		while not isinstance (parent_cat, RootItem):
			if parent_cat.get_id() == cat:
				return True
		return False
	
	def get_data_type (self):
		"""
		Get the data type of this item
		"""
		return self._props["DataType"]
	
	def set_data_type (self, data_type):
		"""
		Set the data type of this item. The data_type argument should be a string.
		Allowed data types are, 'boolean', 'date', 'string', 'float', and 'int'.
		"""
		if not data_type in Field.DATA_TYPES:
			raise OntologyError ("Invalid data type '%s'" % data_type)
		self._props["DataType"] = data_type
		
class FieldMapper (Item):
	def __init__ (self, item_id, item_type, external=False):
		Item.__init__ (self, item_id, item_type, external=external)
		self._fields = {}
	
	def add_field (self, field, notify_child=True):
		if not isinstance (field, Field):
			xesam.error ("Trying to add non-Field type '%s' to %s" % (type(field), self.get_id()))
		
		self._fields[field.get_id()] = field
		if notify_child:
			field.add_category (self, notify_parent=False)
	
	def get_fields (self):
		return self._fields.itervalues()
	
	def get_all_fields (self):
		"""
		Recursively fetch names of all fields implied by this FieldMapper.
		This will not traverse into external branches
		"""
		children = list(self.get_fields())
		for parent in self.get_parents():
			if isinstance(parent, RootItem) or parent.is_external():
				continue
			children.extend(parent.get_all_fields())
		return children
	
	def has_field (self, field):
		if isinstance (field, Field):
			field = field.get_id()
			
		mapper = self
		while mapper:
			for test in mapper.get_fields():
				if field == test.get_id():
					return True
			mapper = mapper.get_parent()
		return False

class Content (FieldMapper):
	def __init__ (self, item_id, external=False):
		FieldMapper.__init__ (self, item_id, item_type=Item.TYPE_CONTENT, external=external)		
		
class Source (FieldMapper):
	def __init__ (self, item_id, external=False):
		FieldMapper.__init__ (self, item_id, item_type=Item.TYPE_SOURCE, external=external)

class KeyFileParseError (Exception):
	def __init__ (self, msg=None):
		if msg is None:
			Exception.__init__ (self)
		else:
			Exception.__init__ (self, msg)

class KeyFileParser :
	"""
	SAX inspired parser for key files (also known as .ini-files).
	Just subclass an implement got_group and got_key methods
	and run parse(file_name).
	"""
	def __init__ (self):
		self._has_group = False
	
	def parse (self, filename):
		line_num = 0
		for line in file(filename).readlines():
				
			line_num += 1
			line = line.strip()
			
			if len (line) == 0:
				continue
			
			if line[0] == "[":
				if not line[-1] == "]":
					raise KeyFileParseError("%s : No closing bracket ']' for group header '%s' at line %s" % (filename, line[1:], line_num))
				self.got_group (line[1:-1])
				self._has_group = True
				continue
			
			if not self._has_group:
				raise KeyFileParseError ("%s : Expected group header at line %s" % (filename, line_num))
				
			keyval = line.split("=")
			if len(keyval) == 1 :
				raise KeyFileParseError ("%s : Line %s must contain a key=value pair" % (filename, line_num))
			
			self.got_key (keyval[0], "=".join(keyval[1:])) 								
	
	def got_group (self, group):
		"""
		Callback for group headers
		"""
		raise NotImplementedError
	
	def got_key (self, key, value):
		"""
		Callback for key-value pairs inside groups
		"""
		raise NotImplemented


