import xesam.ontology
from xesam.ontology import Field, Content, Source, Item, RootItem

import xml.dom.minidom

# Namespaces
RDF = "http://www.w3.org/1999/02/22-rdf-syntax-ns#"
RDFS = "http://www.w3.org/2000/01/rdf-schema#"
DC = "http://purl.org/dc/elements/1.1/"
XESAM = "http://freedesktop.org/standards/xesam/1.0/core#"
XESAMS = "http://freedesktop.org/standards/xesam/1.0/schema#"
XMLS = "http://www.w3.org/2001/XMLSchema#"

def get_label_from_uri (uri):
	if uri.startswith (XESAM):
		return "xesam:"+uri[len(XESAM):]
	else:
		return uri

class RDFXMLParserError (Exception):
	def __init__ (self, msg=None):
		if msg is None:
			Exception.__init__ (self)
		else:
			Exception.__init__ (self, msg)

class RdfXmlParser :
	
	def __init__ (self):
		self._roots = []
		self._fields = {}
		self._cats = {}
		self._external_items = {}
		self._dom = None
		
	def parse (self, rdfs_files, external_ontos=None):
		"""
		The root elements of the ontology
		
		You can pass either a string or file-like object to this method
		"""
		self._external_ontos = external_ontos
		
		for f in rdfs_files:
			xesam.debug ("Parsing %s" % f)
			self._dom = xml.dom.minidom.parse(f)
			self._calc_fields()
			self._calc_categories ()
			self._dom.unlink() # Free memory
						
		self._calc_relations ()
		self._calc_roots()
		return self._roots
	
	def get_roots (self):
		return self._roots
	
	def get_external_refs (self):
		return self._external_items
		
	def _calc_fields (self):
		for prop in self._dom.getElementsByTagNameNS(RDF,"Property"):
			name = prop.getAttributeNS(RDF, "about")
			card = prop.getAttributeNS(XESAMS, "maxCardinality")
			label = self._get_child_char_data ("label", prop, name)
			comment = self._get_child_char_data ("comment", prop, name)
			cats = self._get_child_attributes ("domain", "resource", prop, name)
			parents = self._get_child_attributes ("subPropertyOf", "resource", prop, name)
			data_type = self._get_child_attribute ("range", "resource", prop, name)
			
			if data_type:
				data_type = data_type.replace(XMLS, "").replace(XESAMS, "")
			
			item_props = {"Comment" : comment,
				      "Categories" : cats,
				      "Url" : name,
				      "ParentNames" : parents,
				      "DataType" : data_type,
				      "Cardinality" : card}
			self._fields[name] = Field (label)
			self._fields[name].add_properties(item_props)
		
	def _calc_categories (self):
		for prop in self._dom.getElementsByTagNameNS(RDFS,"Class"):
			name = prop.getAttributeNS(RDF, "about")
			label = self._get_child_char_data ("label", prop, name)
			comment = self._get_child_char_data ("comment", prop, name)
			parents = self._get_child_attributes ("subClassOf", "resource", prop, name)
			item_props = {"Comment" : comment,
				      "Url" : name,
				      "ParentNames" : parents}
			self._cats[name] = Content (label)
			self._cats[name].add_properties(item_props)

	def _get_child_char_data (self, elt_name, node, parent_item):
		"""
		Extract character data from a given child of a node
		@param elt_name: the name of the child to retrieve character data from
		@param node: the node containing the child with elt_name
		@param parent_item: the name of the 'node' argument. This will be the value of it's rdf:about attribute 
		"""
		elts = node.getElementsByTagNameNS(RDFS,elt_name)
		if (elts is None) or (len(elts) == 0):
			#xesam.error ("No child %s for item %s" % (elt_name, parent_item))
			return None
		if len (elts) > 1:
			xesam.error ("Item %s has %s %s children. Expected exactly one. Using first child." % (parent_item, len(elts), elt_name))
		label = elts[0].firstChild.data.encode("utf-8")
		return label
	
	def _get_child_attribute (self, elt_name, attr, node, parent_item):
		"""
		Extract the value of an attribute from a given child of a node
		@param elt_name: the name of the child to retrieve attribute data from. This is assumed to be in the RDFS namespace
		@param attr: the name of the attribute to retrieve. This is assumed to be in the RDF namespace
		@param node: the node containing the child with elt_name
		@param parent_item: the name of the 'node' argument. This will be the value of it's rdf:about attribute 
		"""
		elts = node.getElementsByTagNameNS(RDFS,elt_name)
		if (elts is None) or (len(elts) == 0):
			#xesam.error ("No child %s for item %s" % (elt_name, parent_item))
			return None
		if len (elts) > 1:
			xesam.error ("Item %s has %s %s children. Expected exactly one. Using first child." % (parent_item, len(elts), elt_name))
		val = elts[0].getAttributeNS(RDF, attr)
		return val
	
	def _get_child_attributes (self, elt_name, attr, node, parent_item):
		"""
		Extract the values of an attribute from all children with a given name
		@param elt_name: the name of the child to retrieve attribute data from. This is assumed to be in the RDFS namespace
		@param attr: the name of the attribute to retrieve. This is assumed to be in the RDF namespace
		@param node: the node containing the child with elt_name
		@param parent_item: the name of the 'node' argument. This will be the value of it's rdf:about attribute 
		"""
		elts = node.getElementsByTagNameNS(RDFS,elt_name)
		if (elts is None) or (len(elts) == 0):
			#xesam.error ("No child %s for item %s" % (elt_name, parent_item))
			return None
		values = []
		for val in elts:
			values.append(val.getAttributeNS(RDF, attr))
		return values
	
	def _calc_relations (self):
		# Compile parent-child relations of fields and categories
		for item_type in self._fields, self._cats:
			for name, item in item_type.iteritems() :
				if item["ParentNames"] != None :
					for parent in item["ParentNames"]:
						if isinstance(parent,RootItem) or parent is None:
							continue
						try:
							parent = item_type[parent]
						except KeyError:
							try:
								parent = self._resolve_external (parent)
							except KeyError:
								xesam.error ("%s %s references unknown parent %s" % (self._get_type_desc(item), name, parent))
								continue
						item.add_parent (parent)
			
			# Resolve the item types only after the hierarchy has been calculated
			# since we resolve via tree traversal
			for name, item in item_type.iteritems() :
					if item.is_child_of (xesam.ontology.XESAM_URL_SOURCE):
						item.set_type (Item.TYPE_SOURCE)
		
		# Add fields to categories
		for name, item in self._fields.iteritems():
			try:
				# Raise error if "Categories" is unset or has length 0
				if len(item["Categories"]) == 0:
					raise KeyError ("Categories")
			except KeyError:
				xesam.error ("Item %s does not have a category" % item.get_id())
				continue
			for cat_name in item["Categories"]:
				cat_item = None
				if self._cats.has_key(cat_name):
					cat_item = self._cats[cat_name]
				else:
					try:
						cat_item = self._resolve_external (cat_name)
						self._cats[cat_name] = cat_item
					except KeyError, e:
						xesam.error ("In field '%s': Unresolved reference to '%s'" % (name, cat_name))
						continue
				cat_item.add_field (item) # this also adds the cat of the field to the real Category obj
		
		# Add external entities to the ontology
		for item in self._external_items.itervalues():
			self._cats[item.get_id()] = item
	
	def _resolve_external (self, item_id):
		# If we are given an URI, resolve the label
		item_id = get_label_from_uri (item_id)
		
		# Check if the given item is already cached
		if self._external_items.has_key(item_id):
			return self._external_items[item_id]
		
		# Look it up in the external ontos, we have not already encountered it
		if not self._external_ontos:
			raise KeyError(item_id)
		else:
			xesam.debug ("Resolving external reference %s" % item_id)
			for onto in self._external_ontos:
				# Create a dummy category for the externally linked category
				if onto.has_content(item_id) or \
				   onto.has_source(item_id)  or \
				   onto.has_field(item_id):
					self._external_items[item_id] = onto.get_item(item_id)
					self._external_items[item_id].set_external(True)
					return self._external_items[item_id]
				else:
					raise KeyError (item_id)

	def _calc_roots (self):
		for name, item in self._cats.iteritems():
			all_parents_are_roots = True
			for parent in item.get_parents():
				if( not isinstance (parent, xesam.ontology.RootItem)) and \
				   (not parent.is_external()):
					all_parents_are_roots = False
			if all_parents_are_roots and not item.is_external():
				self._roots.append(item)
		
		# Also include fields without a cat and fields in external cats
		# as roots
		for name, field in self._fields.iteritems():
			has_valid_parent = True
			for parent in list(field.get_parents()) + field.get_categories():
				if (isinstance (parent, xesam.ontology.RootItem)) or \
				    parent.is_external():
					has_valid_parent = False
			if (not has_valid_parent) and (not field.is_external()):
				self._roots.append(field)
			
	def _get_type_desc (self, item):
		if item.get_type () == Item.TYPE_FIELD: return "Field"
		elif item.get_type () == Item.TYPE_CONTENT: return "Content"
		elif item.get_type () == Item.TYPE_SOURCE: return "Source"
		else: 
			xesam.error ("Unknown type of item " + str(item))
			return "Unknown Type"
		
